// -------------------------------------------------------------------------------------
//
// FishFa30.CPP                                                   Ulrich Mller 19.07.02
//                                                                PCS1700 C++Builder 4.0
//                                                                Version         3.0.23
// FishFa30.Unit : Basisklasse zur Ansteuerung der fischertechnik Interfaces
// aus C++Builder. Die Funktionalitt entspricht der von umFish20Ex.
// Neu sind die Geschwindigkeitssteuerung und der RobMode (ansync. Anfahren
// einer Position).
//
//--------------------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "FishFa30.h"

#pragma package(smart_init)

// --- TFishFaceException : Constructor -------------------------------------

TFishFaceException::TFishFaceException(int nr, char* text)
{
    Fnr = nr;
    Ftext = text;
}

// --- TFishFace : Constructor / Destructor ---------------------------------

TFishFace::TFishFace()
{
    LoadFish();
    ft.AnalogScan = 0;
    ft.ECount = 8;
    ft.LPTDelay = 0;
    ft.LPTAnalog = 0;
    ft.PollInterval = 0;
}
TFishFace::TFishFace(bool AnalogScan, bool Slave, unsigned int PollInterval)
{
    LoadFish();
    ft.AnalogScan = AnalogScan ? 1 : 0;
    ft.ECount = Slave ? 16 : 8;
    ft.LPTDelay = 0;
    ft.LPTAnalog = 0;
    ft.PollInterval = PollInterval;
}
TFishFace::TFishFace(bool AnalogScan, bool Slave, unsigned int PollInterval,
    unsigned int LPTAnalog, unsigned int LPTDelay)
{
    LoadFish();
    ft.AnalogScan = AnalogScan ? 1 : 0;
    ft.ECount = Slave ? 16 : 8;
    ft.LPTDelay = LPTDelay;
    ft.LPTAnalog = LPTAnalog;
    ft.PollInterval = PollInterval;
}
TFishFace::~TFishFace()
{
    FreeFish();
}

// --- FreeFish : Freigeben umFish30.DLL ---------------------------------------

void TFishFace::FreeFish()
{
	FreeLibrary(umFishLib);
	umFishLib = NULL;
	return;
}

// --- LoadFish : Laderoutine umFish30.DLL -------------------------------------

BOOLEAN TFishFace::LoadFish(void)
{
    umFishLib = LoadLibrary("umFish30.DLL");
	if(!umFishLib) return FALSE;

	umOpenInterface   = (TOpenInterface)  GetProcAddress(umFishLib, "umOpenInterface");
	umCloseInterface  = (TCloseInterface) GetProcAddress(umFishLib, "umCloseInterface");
	umGetVersion      = (TVersion)        GetProcAddress(umFishLib, "umGetVersion");
    umGetAnalogDirect = (TGetAnalogDirect)GetProcAddress(umFishLib, "umGetAnalogDirect");
	umGetInput        = (TGetInput)       GetProcAddress(umFishLib, "umGetInput");
	umSetLamp         = (TSetLamp)        GetProcAddress(umFishLib, "umSetLamp");
	umSetMotor        = (TSetMotor)       GetProcAddress(umFishLib, "umSetMotor");
	umSetMotorEx      = (TSetMotorEx)     GetProcAddress(umFishLib, "umSetMotorEx");
	umSetMotors       = (TSetMotors)      GetProcAddress(umFishLib, "umSetMotors");
 	umSetMotorsEx     = (TSetMotorsEx)    GetProcAddress(umFishLib, "umSetMotorsEx");
	umRobMotors       = (TRobMotors)      GetProcAddress(umFishLib, "umRobMotors");
	umRobMotor        = (TRobMotor)       GetProcAddress(umFishLib, "umRobMotor");

	umStartDriver     = (TStartDriver)    GetProcAddress(umFishLib, "umStartDriver");
	umStopDriver      = (TStopDriver)     GetProcAddress(umFishLib, "umStopDriver");
	return TRUE;
}

// --- PollInterval : GetPollInterval -------------------------------------

unsigned int __fastcall TFishFace::GetPollInterval()
{
    return ft.PollInterval;
}

// --- AnalogScan : GetAnalogScan --------------------------------------------

bool __fastcall TFishFace::GetAnalogScan()
{
    return ft.AnalogScan ? true : false;
}

// --- AnalogsEX : GetAnalogsEX ----------------------------------------------

unsigned int __fastcall TFishFace::GetAnalogsEX()
{
    return ft.Analogs[0];
}

// --- AnalogsEY : GetAnalogsEY -----------------------------------------------

unsigned int __fastcall TFishFace::GetAnalogsEY()
{
    return ft.Analogs[1];
}

// --- Inputs : GetInputsP -----------------------------------------------------

unsigned int __fastcall TFishFace::GetInputsP()
{
    return ft.InputStatus;
}

// --- Outputs : GetOutputs ----------------------------------------------------

unsigned int __fastcall TFishFace::GetOutputs()
{
    return ft.MotorStatus;
}

// --- ClearCounter : Lschen eines Counters ----------------------------------

void TFishFace::ClearCounter(unsigned int InputNr)
{
    ft.Counters[InputNr-1] = 0;
}

// --- ClearCounters : Lschen aller Counter ----------------------------------

void TFishFace::ClearCounters()
{
    for(int i=0; i<16; i++) ft.Counters[i] = 0;
}

// --- ClearMotors : Lschen aller M-Ausgnge --------------------------------

void TFishFace::ClearMotors()
{
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.ClearMotors");
    if(umSetMotors(ft, 0) == ftiFehler)
        throw TFishFaceException(1, "InterfaceProblem.ClearMotors");
}

// --- CloseInterface : Schlieen Interface ----------------------------------

void TFishFace::CloseInterface()
{
    umCloseInterface(ft);
    FOpen = false;
}

// --- Finish : Feststellen Endewunsch (NotHalt, Escape, InputNr) ------------

bool TFishFace::Finish()
{
    return Finish(0);
}
bool TFishFace::Finish(unsigned int InputNr)
{
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.Finish");
    if(ft.InputStatus == ftiFehler)
        throw TFishFaceException(1, "InterfaceProblem.Finish");
    if(FMessages) Application->ProcessMessages();
    if(FNotHalt || GetAsyncKeyState(VK_ESCAPE) !=0) return true;
    if(InputNr && GetInput(InputNr)) return true;
    return false;
}

// --- GetAnalog : Auslesen Analogwert ---------------------------------------

unsigned int TFishFace::GetAnalog(unsigned int AnalogNr)
{
    if(FMessages) Application->ProcessMessages();
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.GetAnalog");
    if(ft.Analogs[AnalogNr] == ftiFehler)
        throw TFishFaceException(1, "InterfaceProblem.GetAnalog");
    return ft.Analogs[AnalogNr];
}

// --- GetAnalogDirect : Direktes Auslesen der Analogwerte --------------------

unsigned int TFishFace::GetAnalogDirect(unsigned int AnalogNr)
{
unsigned int res;

    if(FMessages) Application->ProcessMessages();
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.GetAnalogDirect");
    res = umGetAnalogDirect(ft, AnalogNr);
    if(res == ftiFehler)
        throw TFishFaceException(1, "InterfaceProblem.GetAnalogDirect");
    return res;
}

// --- GetCounter : Lesen eines ImpulsCounters --------------------------------

unsigned int TFishFace::GetCounter(unsigned int InputNr)
{
    return ft.Counters[InputNr-1];
}

// --- GetInput : Bestimmen Status E-Eingang ----------------------------------

bool TFishFace::GetInput(unsigned int InputNr)
{
    if(FMessages) Application->ProcessMessages();
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.GetInput");
    int res = umGetInput(ft, InputNr);
    if(ft.InputStatus == ftiFehler)
        throw TFishFaceException(1, "InterfaceProblem.GetInput");
    return res;
}

// --- GetInputs : Auslesen Status E-Eingnge ---------------------------------

unsigned int TFishFace::GetInputs()
{
    if(FMessages) Application->ProcessMessages();
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.GetInputs");
    if(ft.InputStatus == ftiFehler)
        throw TFishFaceException(1, "InterfaceProblem.GetInputs");
    return ft.InputStatus;
}

// --- LPTAnalog : GetLPTAnalog -------------------------------------------

unsigned int __fastcall TFishFace::GetLPTAnalog()
{
    return ft.LPTDelay;
}

// --- LPTDelay : GetLPTDelay ---------------------------------------------

unsigned int __fastcall TFishFace::GetLPTDelay()
{
    return ft.LPTDelay;
}

// --- OpenInterface : Verbindung zum Interface herstellen ---------------------

void TFishFace::OpenInterface(LPCSTR PortName, bool Messages)
{
    if(umOpenInterface(ft, PortName) == ftiFehler)
        throw TFishFaceException(1, "InterfaceProblem.OpenInterface");
    FMessages = Messages;
    FOpen     = true;
    FNotHalt  = false;
    ClearMotors();
}

// --- Pause : Programm anhalten ----------------------------------------------

void TFishFace::Pause(unsigned int mSek)
{
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.Pause");
    unsigned int sZeit = GetTickCount();
    do
    {
        if(FMessages) Application->ProcessMessages();
        if(ft.InputStatus == ftiFehler)
            throw TFishFaceException(1, "InterfaceProblem.Pause");
        if(FNotHalt || GetAsyncKeyState(VK_ESCAPE) !=0) return;
        Sleep(ft.PollInterval);
    } while(GetTickCount() - sZeit < mSek);
}

// --- SetCounter : Setzen eines ImpulsCounters -------------------------------

void TFishFace::SetCounter(unsigned int InputNr, unsigned int Value)
{
    ft.Counters[InputNr-1] = Value;
}

// --- SetLamp : Setzen eines 'halben' M-Ausganges ----------------------------

void TFishFace::SetLamp(unsigned int LampNr, bool OnOff)
{
    if(FMessages) Application->ProcessMessages();
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.SetLamp");
    if(umSetLamp(ft, LampNr, OnOff ? 1 : 0) == ftiFehler)
        throw TFishFaceException(1, "InterfaceProblem.SetLamp");
}


// --- SetMotor : Setzen eines M-Ausganges -------------------------------------

void TFishFace::SetMotor(unsigned int MotorNr, unsigned int Direction)
{
    if(FMessages) Application->ProcessMessages();
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.SetMotor");
    if(umSetMotor(ft, MotorNr, Direction) == ftiFehler)
        throw TFishFaceException(1, "InterfaceProblem.SetMotor");
}
void TFishFace::SetMotor(unsigned int MotorNr, unsigned int Direction, unsigned int Speed)
{
    if(FMessages) Application->ProcessMessages();
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.SetMotor");
    if(umSetMotorEx(ft, MotorNr, Direction, Speed) == ftiFehler)
        throw TFishFaceException(1, "InterfaceProblem.SetMotor");
}
void TFishFace::SetMotor(unsigned int MotorNr, unsigned int Direction,
        unsigned int Speed, unsigned int Counter)
{
    if(FMessages) Application->ProcessMessages();
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.SetMotor");
    if(umRobMotor(ft, MotorNr, Direction, Speed, Counter) == ftiFehler)
        throw TFishFaceException(1, "InterfaceProblem.SetMotor");
}

// --- SetMotors : Setzen aller M-Ausgnge ------------------------------------

void TFishFace::SetMotors(unsigned int MotorStatus, unsigned int SpeedStatus,
        unsigned int ModeStatus)
{
    if(FMessages) Application->ProcessMessages();
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.SetMotors");
    if(umRobMotors(ft, MotorStatus, SpeedStatus, ModeStatus) == ftiFehler)
        throw TFishFaceException(1, "InterfaceProblem.SetMotors");
}
void TFishFace::SetMotors(unsigned int MotorStatus, unsigned int SpeedStatus)
{
    if(FMessages) Application->ProcessMessages();
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.SetMotors");
    if(umSetMotorsEx(ft, MotorStatus, SpeedStatus) == ftiFehler)
        throw TFishFaceException(1, "InterfaceProblem.SetMotors");
}
void TFishFace::SetMotors(unsigned int MotorStatus)
{
    if(FMessages) Application->ProcessMessages();
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.SetMotors");
    if(umSetMotors(ft, MotorStatus) == ftiFehler)
        throw TFishFaceException(1, "InterfaceProblem.SetMotors");
}

// --- Slave : GetSlave --------------------------------------------------

bool __fastcall TFishFace::GetSlave()
{
    return (ft.ECount == 16)? true : false;
}

// --- Version : GetVersion --------------------------------------------------

unsigned int __fastcall TFishFace::GetVersion()
{
  return 30023;
}

// --- WaitForChange : warten auf NrOfChanges Impulse von InputNr ------------

void TFishFace::WaitForChange(unsigned int InputNr, unsigned int NrOfChanges,
        unsigned int TermInputNr)
{
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.WaitForChange");
    ft.Counters[InputNr-1] = 0;
    while(ft.Counters[InputNr-1] < NrOfChanges)
    {
        if(ft.InputStatus == ftiFehler)
            throw TFishFaceException(1, "InterfaceProblem.WaitForChange");
        if(FNotHalt || GetAsyncKeyState(VK_ESCAPE)) return;
        if(TermInputNr > 0) if(umGetInput(ft, InputNr)) return;
        if(FMessages) Application->ProcessMessages();
    }
}
void TFishFace::WaitForChange(unsigned int InputNr, unsigned int NrOfChanges)
{
    WaitForChange(InputNr, NrOfChanges, 0);
}

// --- WaitForHigh : Warten auf einen fals/true-Durchgang in InputNr ---------

void TFishFace::WaitForHigh(unsigned int InputNr)
{
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.WaitForHigh");
    WaitForInput(InputNr, false);
    WaitForInput(InputNr, true);
}

// --- WaitForInput : Warten auf InputNr = OnOff ------------------------------

void TFishFace::WaitForInput(unsigned int InputNr, bool OnOff)
{
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.WaitForInput");
    do
    {
        if(ft.InputStatus == ftiFehler)
            throw TFishFaceException(1, "InterfaceProblem.WaitForInput");
        if(FNotHalt || GetAsyncKeyState(VK_ESCAPE)) return;
        if(FMessages) Application->ProcessMessages();
    } while(!umGetInput(ft, InputNr) == OnOff);
}

// --- WaitForLow : Warten auf einen true/false-Durchgang in InputNr ----------

void TFishFace::WaitForLow(unsigned int InputNr)
{
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.WaitForLow");
    WaitForInput(InputNr, true);
    WaitForInput(InputNr, false);
}

// WaitForMotors : Warten auf MotorReady-Ereignis oder Ablauf mSek -------------

unsigned int TFishFace::WaitForMotors(unsigned int mSek,
    unsigned int M1, unsigned int M2, unsigned int M3, unsigned int M4,
    unsigned int M5, unsigned int M6, unsigned int M7, unsigned int M8)
{
    const unsigned int mM3[9] = {0, 0x3, 0xC, 0x30, 0xC0, 0x300, 0xC00, 0x3000, 0xC000};
    unsigned int c, t;
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.WaitForMotors");
    t = GetTickCount() + mSek;
    do
    {
        c = 0;
        if(M1) c += ft.MotorStatus & mM3[M1];
        if(M2) c += ft.MotorStatus & mM3[M2];
        if(M3) c += ft.MotorStatus & mM3[M3];
        if(M4) c += ft.MotorStatus & mM3[M4];
        if(M5) c += ft.MotorStatus & mM3[M5];
        if(M6) c += ft.MotorStatus & mM3[M6];
        if(M7) c += ft.MotorStatus & mM3[M7];
        if(M8) c += ft.MotorStatus & mM3[M8];

        if(!c) return ftiEnde;

        // --- Abschalten der Motoren der Parameterliste --------------
        if(FNotHalt || GetAsyncKeyState(VK_ESCAPE))
        {
            if(M1) umSetMotor(ft, M1, 0);
            if(M2) umSetMotor(ft, M2, 0);
            if(M3) umSetMotor(ft, M3, 0);
            if(M4) umSetMotor(ft, M4, 0);
            if(M5) umSetMotor(ft, M5, 0);
            if(M6) umSetMotor(ft, M6, 0);
            if(M7) umSetMotor(ft, M7, 0);
            if(M8) umSetMotor(ft, M8, 0);
        }

        if(ft.InputStatus == ftiFehler)
            throw TFishFaceException(1, "InterfaceProblem.WaitForMotors");
        if(FNotHalt) return ftiNotHalt;
        if(GetAsyncKeyState(VK_ESCAPE)) return ftiESC;
        if(FMessages) Application->ProcessMessages();
        Sleep(ft.PollInterval);
    } while(!mSek || (t>GetTickCount()));
    return ftiTime;
}

// --- WaitForPositionDown : Position = Counter - InputNr-Impulse -------------

void TFishFace::WaitForPositionDown(unsigned int InputNr, unsigned int & Counter,
        unsigned int Position, unsigned int TermInputNr)
{
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.WaitForPositionDown");
    WaitForChange(InputNr, Counter-Position, TermInputNr);
    Counter -= ft.Counters[InputNr];
}

// --- WaitForPositionUp : Position = Counter + InputNr-Impulse ---------------

void TFishFace::WaitForPositionUp(unsigned int InputNr, unsigned int & Counter,
        unsigned int Position, unsigned int TermInputNr)
{
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.WaitForPositionUp");
    WaitForChange(InputNr, Position-Counter, TermInputNr);
    Counter += ft.Counters[InputNr];
}

// --- WaitForTime : Programm anhalten ----------------------------------------

void TFishFace::WaitForTime(unsigned int mSek)
{
    if(!FOpen) throw TFishFaceException(2, "KeinOpen.WaitForTime");
    Pause(mSek);
}




