// -----------------------------------------------------------------------------------------
//
// umFish30.CPP                                                       Ulrich Mller 17.05.02
//                                                                    PCS1700       VC++ 6.0
// Basisfunktionen fr das fischertechnik Interface (seriell und parallel)
// Der Interfacestatus wird asynchron ermittelt(E- und A-Eingnge) bzw. gesetzt (M-Ausgnge)
// Die Impulse an den E-Ausgngen werden in Counters gezhlt.
// zustzlich wird die Motorgeschwindigkeit kontrolliert (MotorStatus != 0).
// Im RobModus (ModeStatus = 1) wird bei Erreichen der Zielposition der zugehrende
// Motor abgeschaltet (Counter = 0 oder EndTaster true).
// umOpenInterface startet dazu den MultiMediaTimer mit PollInterface als CallBack-Routine
// in einem eigenen Thread
// 
// -----------------------------------------------------------------------------------------

#include <windows.h>
#include <mmsystem.h>
#include <ole2.h>
#include "umFish30.H"


DWORD OpenInterfaceCOM(ftiDCB &Interface, LPCSTR ComName);
DWORD OpenInterfaceLPT(ftiDCB &Interface, LPCSTR LptName);
DWORD GetInputs(ftiDCB &Interface);
DWORD GetInputsCOM1(ftiDCB &Interface);
DWORD GetInputsCOM2(ftiDCB &Interface);
DWORD GetInputsLPT(ftiDCB &Interface);
DWORD GetAnalog(ftiDCB &Interface, DWORD AnalogNr);
DWORD GetAnalogCOM1(ftiDCB &Interface, DWORD AnalogNr);
DWORD GetAnalogCOM2(ftiDCB &Interface, DWORD AnalogNr);
DWORD GetAnalogLPT(ftiDCB &Interface, DWORD AnalogNr);
DWORD GetLPT(ftiDCB &Interface);
VOID  SetLPT(ftiDCB &Interface, WORD Wert);


const FVersion = 322;
ftiDCB ftiSave;
ftiDCB ftDCB[8];
int    DCBPool[8];


// PollInterface ---------------------------------------------------------------------------
//
// Zentrales Pollen eines Interfaces, setzen InputStatus, OutputStatus, Analogs
// Fhren eines Impulszhlers (Counters) fr die E-Eingnge
// Geschwindigkeitssteuerung der M-Ausgnge, MotorStatus hat Prioritt fr Motor Aus
// Die A-Eingnge werden nur bei AnalogScan > 0 gepollt und dann im Wechsel 0/1
// Beim seriellen Interface wird dann der IntputStatus ber das AnalogLesen ermittelt
//
// -----------------------------------------------------------------------------------------

void CALLBACK PollInterface(UINT wTimerID, UINT msg, DWORD DCB, DWORD dw1, DWORD dw2)
{
	DWORD StatusNeu, StatusAlt, StatusDelta, iMot, iEnd, iImpuls; 
	DWORD SpeedNr, SpeedValue, SollDirection, AktDirection;

	ftiDCB *d = (ftiDCB*)DCB;

	// --- Refresh OutputStatus, get InputStatus, Analogs -------------------------------
	// 
	// Analogs wird bei Bedarf (AnalogScan > 0) wechselnd von EX und EY gelesen
	// bei COM zusammen mit InputStatus, bei LPT separat
	//

	StatusAlt = d->InputStatus;
	if(StatusAlt == ftiFehler) return;
	if(d->PortID == 0) 
	{   // --- COM-Interface
		if(d->AnalogScan == 0) StatusNeu = GetInputs(*d);
		else 
		{
			d->Analogs[d->AnalogScan-1] = GetAnalog(*d, d->AnalogScan-1);
			StatusNeu = d->InputStatus;
			if(d->AnalogScan == 1) d->AnalogScan = 2; else d->AnalogScan = 1;
		}
	}
	else 
	{   // --- LPT-Interface
		StatusNeu = GetInputs(*d);
		if(d->AnalogScan > 0) 
		{
			d->Analogs[d->AnalogScan-1] = GetAnalog(*d, d->AnalogScan-1);
			if(d->AnalogScan == 1) d->AnalogScan = 2; else d->AnalogScan = 1;
		}
	}

	StatusDelta = StatusAlt ^ StatusNeu;
	d->InputStatus = StatusNeu;


	// ---- Schleife ber Motoren ---------------------------------------

	d->TaktNr = (d->TaktNr > 14) ? 0 : d->TaktNr+1;
	for (iMot=1; iMot<=d->ECount/2; iMot++)
	{
		iEnd    = (iMot-1)*2+1;
		iImpuls = (iMot-1)*2+2;
		int MotMode = (d->ModeStatus & SMaske[iMot]);
		if(MotMode == 0)
		{   // --- NormalMode : incr Impulstaster
			if((StatusDelta & EMaske[iEnd]) > 0)    d->Counters[iEnd-1]++;
			if((StatusDelta & EMaske[iImpuls]) > 0) d->Counters[iImpuls-1]++;
		}
		else
		{   // --- RobMode : Zhlen Impulse, Abschalten bei 0, iEnd(ftiLinks)
			if((StatusDelta & EMaske[iImpuls]) > 0) d->Counters[iImpuls-1]--;
			if(d->Counters[iImpuls-1] <= 6)
			{   // --- Am Ende der Strecke Bremsen
				d->SpeedStatus &= ~SMaske[iMot];
				d->SpeedStatus |= 0x8 << (iMot-1)*4;
			}
			int MotDirection = (d->MotorStatus & ~MAus[iMot]) >> (iMot-1)*2;
			BOOL MotEnde = (d->InputStatus & EMaske[iEnd]) ? true : false;
			if((d->Counters[iImpuls-1] == 0) || ((MotDirection == ftiLinks) && MotEnde))
			{
				d->MotorStatus &= MAus[iMot];
				d->OutputStatus = d->MotorStatus;
				d->ModeStatus &= ~SMaske[iMot];
				d->Counters[iImpuls-1] = 0;
			}
		}
		
		// ---- Geschwindigkeitssteuerung -----------

		SollDirection = d->MotorStatus & ~MAus[iMot];
		if (SollDirection > 0)
		{
			SpeedNr = (d->SpeedStatus & SMaske[iMot]) >> (iMot-1)*4;
			if((SpeedNr < 15) && (SpeedNr > 0))
			{
				d->OutputStatus &= MAus[iMot];     // --- Motor aus
				SpeedValue = ((OnOffTab[SpeedNr] & EMaske[d->TaktNr]) > 0 ? 3 : 0)
					<< (iMot-1)*2;
				AktDirection = SollDirection & SpeedValue ;
				d->OutputStatus |= AktDirection;   // --- Motor Links, Rechts, Aus
			}
		}

	}
}


// umGetVersion ----------------------------------------------------------------------------
//
DWORD __stdcall umGetVersion(){return(FVersion);}




// umOpenInterface ------------------------------------------------------------
//
// Setzen Ausgangswerte, Open Schnittstelle
//  PortName (ComName)  COM1 - 9 : alle
//                      LPT1 - 3 (Win9x, wenn Standard Port, CreateFile)
//                      LPT (Win9x) PortID zur Identification, kein CreateFile
//                          WinNT in Registry eingestellter Port, 
//                          WinRT.SYS mu laufen (Allow Conflicts reicht)
//
// --------------------------------------------------------------------------

DWORD __stdcall umOpenInterface(ftiDCB &Interface, LPCSTR ComName)
{
	DWORD res;

	res = ftiFehler;
	Interface.hCom = 0;
	Interface.OutputStatus = 0;
	Interface.MotorStatus = 0;
	Interface.SpeedStatus = 0;
	Interface.ModeStatus = 0;
	Interface.TaktNr = 0;
	Interface.InputStatus = 0;
	for(int i=0; i<16; i++) Interface.Counters[i] = 0;
	if(ComName == 0) return(ftiFehler);

	if(ComName[0] == 'C' || ComName[0] == 'c')
		res = OpenInterfaceCOM(Interface, ComName);
	else
		res = OpenInterfaceLPT(Interface, ComName);
	if(res == ftiFehler) {ftiSave = Interface; return(ftiFehler);}

	if(Interface.AnalogScan != 1) Interface.AnalogScan = 0;
	if(Interface.ECount != 16) Interface.ECount = 8;
	if(Interface.PortID == 0) 
	{   // --- serielles Interface
		if(Interface.PollInterval < 10) Interface.PollInterval = 10;
		if((Interface.ECount == 16) && (Interface.PollInterval < 12))
			Interface.PollInterval = 12;
		if((Interface.AnalogScan > 0) && (Interface.PollInterval < 16))
			Interface.PollInterval = 16;	
	}
	else 
	{                                          // --- paralleles Interface 
		if(Interface.PollInterval < 1) Interface.PollInterval = 1;
		if((Interface.AnalogScan > 0) && (Interface.PollInterval < 20))
			Interface.PollInterval = 100; 
	}


// ------- Start MultiMediaTimer -------------------------------------------------------

	Interface.FID = timeSetEvent(Interface.PollInterval,
	                            0,                     // --- Auflsung in mSek (0 = max)
	                            PollInterface,         // --- CallBack-Routine
								DWORD(&Interface),     // --- bergabe ftiDCB 
								TIME_PERIODIC); 
	ftiSave = Interface;
	return(res);
}

// OpenInterfaceCOM ------------------------------------------------------------
// 
// Setzen Ausgangswerte, Open Schnittstelle
//
// --------------------------------------------------------------------------

DWORD OpenInterfaceCOM(ftiDCB &Interface, LPCSTR ComName)
{
	DCB dcb;
	COMMTIMEOUTS cTime;
	DWORD res;
	HANDLE hCom;

// --- ffnen des ComPorts ---

	hCom = CreateFile(ComName,
	GENERIC_READ | GENERIC_WRITE,
	0,                               // --- Open exklusiv
	NULL,                            // --- keine Security-Attribute
	OPEN_EXISTING,                   // --- Open Modus
	0,                               // --- kein overlapped I/O
	NULL);                           // --- kein Template
  
	if (hCom == INVALID_HANDLE_VALUE) return(ftiFehler);

// --- Einstellen der COM-Parameter ---

	if (!GetCommState(hCom, &dcb)) return(ftiFehler);

	dcb.BaudRate = 9600;
	dcb.ByteSize = 8;
	dcb.Parity   = NOPARITY;
	dcb.StopBits = ONESTOPBIT;

	if (!SetCommState(hCom, &dcb)) return(ftiFehler);

	cTime.ReadIntervalTimeout         = 5;
	cTime.ReadTotalTimeoutMultiplier  = 4;
	cTime.ReadTotalTimeoutConstant    = 30;
	cTime.WriteTotalTimeoutMultiplier = 1;
	cTime.WriteTotalTimeoutConstant   = 10;

	if (!SetCommTimeouts(hCom, &cTime)) return(ftiFehler);

	Interface.hCom         = hCom;
	Interface.PortID       = 0;
	Interface.LPTDelay     = 1;
	Interface.LPTAnalog    = 1;
	Interface.OutputStatus = 0;
	res = GetInputsCOM1(Interface);
	if (res == ftiFehler) 
	{
		CloseHandle(hCom);
		Interface.hCom = 0;
		return(ftiFehler);
    }

    return(res);
}
 
// OpenInterfaceLPT ------------------------------------------------------------
// 
// Setzen Ausgangswerte, Open LPT-Schnittstelle
// Wenn Version = Wondows NT / 2000 wird ber OpenInterfaceRT der WinRT-Treiber geffnet
// sonst wird bei LPTn direkt auf das Interface zugegriffen
//                     (LPT1 : 0x378, 2 : 0x278, 3 : 0x3BC)
//                LPT  ebenfalls ber OpenInterfaceRT der WinRT-Treiber geffnet
//
// -----------------------------------------------------------------------------

DWORD OpenInterfaceLPT(ftiDCB &Interface, LPCSTR LptName)
{
	DWORD res;
	HANDLE hLPT;

	if(GetVersion() <0x80000000)          // --- Windows NT       
	{
		return(OpenInterfaceRT(Interface));
	}
	else
	{   // --- ffnen LPT-Port (Schutzmanahme, fr Zugriff ber Get/SetLPT nicht erforderl.
		if(LptName[3] != 0) 
		{
			hLPT = CreateFile(LptName,
				GENERIC_READ | GENERIC_WRITE,
				0,                                 // --- Open eklusiv
				NULL,                              // --- keine Security-Attribute
				OPEN_EXISTING,                     // --- Open-Modus
				0,                                 // --- kein Overlapped
				NULL);                             // --- kein Template

			if(hLPT == INVALID_HANDLE_VALUE) return(ftiFehler);

			switch (LptName[3]) 
			{
				case '1': Interface.PortID = 0x378; break; 
				case '2': Interface.PortID = 0x278; break;
				case '3': Interface.PortID = 0x3BC; break;
				default:     Interface.PortID = 0x378;
			}
			Interface.hCom = hLPT;
			if(Interface.LPTDelay == 0) Interface.LPTDelay    = 10;
			if(Interface.LPTAnalog == 0) Interface.LPTAnalog  = 5;
			Interface.OutputStatus = 0;
			res = GetInputs(Interface);
			if(res == ftiFehler) 
			{
				CloseHandle(hLPT);
				Interface.hCom = 0;
				return(ftiFehler);
			}
			return(res);
		}
		else
		{
			return(OpenInterfaceRT(Interface));
		}
	}
}


// umCloseInterface ------------------------------------------------------------
// 
// Schlieen der Schnittstelle
//
// --------------------------------------------------------------------------

DWORD __stdcall umCloseInterface(ftiDCB &Interface)
{
	DWORD res;
	HANDLE hCom = Interface.hCom;
	if(hCom == 0) return(0);


	Interface.ModeStatus   = 0;
	Interface.OutputStatus = 0;              // --- Lschen M-Ausgnge
	Sleep(4 * Interface.PollInterval);       // --- dem MM-Timer Zeitlassen
	if(Interface.FID != 0) timeKillEvent(Interface.FID); // dann Killen
	                                          // --- Freigabe der Handles
	if (Interface.PortID == 1)	res = CloseInterfaceRT(Interface); 
	else res = CloseHandle(hCom);

	Interface.hCom   = 0;                    // --- Open-Flag -----------------
	Interface.PortID = 0;
	ftiSave = Interface;                     // --- fr ein Close beim Detach
	return(res);
}


// GetInputs -----------------------------------------------------------------
// 
// Setzen Motorstatus, Lesen Inputstatus
//
// -----------------------------------------------------------------------------

DWORD GetInputs(ftiDCB &Interface)
{
	if(Interface.hCom == 0) return(ftiFehler);
		if(Interface.PortID == 0)
			if (Interface.ECount == 8) return(GetInputsCOM1(Interface));
		else return(GetInputsCOM2(Interface));
	else
		if(Interface.PortID == 1) return(GetInputsRT(Interface)); 
		else return(GetInputsLPT(Interface));
}

// GetInputsCOM1 -------------------------------------------------------------
//
// Setzen Motorstatus, Lesen Inputstatus
//
// --------------------------------------------------------------------------

DWORD GetInputsCOM1(ftiDCB &Interface)
{
	HANDLE hCom = Interface.hCom;
	unsigned char Befehl[2];
	unsigned char IStatus[1];
	BOOL  res;
	DWORD Geschrieben;
	DWORD Gelesen;

	Befehl[0] = '\xC1';
	Befehl[1] = unsigned char(Interface.OutputStatus);

	res = WriteFile(hCom, Befehl, 2, &Geschrieben, NULL);
	res = ReadFile(hCom, IStatus, 1, &Gelesen, NULL);
	if (Gelesen == 0) {return(ftiFehler);}

	Interface.InputStatus = unsigned char(IStatus[0]);
	return(Interface.InputStatus);
}


// GetInputsCOM2 -------------------------------------------------------------
//
// Setzen Motorstatus, Lesen Inputstatus
//
// --------------------------------------------------------------------------

DWORD GetInputsCOM2(ftiDCB &Interface)
{
	HANDLE hCom = Interface.hCom;
	unsigned char Befehl[3];
	unsigned char IStatus[2];
	BOOL  res;
	DWORD Geschrieben;
	DWORD Gelesen;

	Befehl[0] = '\xC2';
	Befehl[1] = unsigned char(Interface.OutputStatus & 0xFF);
	Befehl[2] = unsigned char((Interface.OutputStatus & 0xFF00)>>8);

	res = WriteFile(hCom, Befehl, 3, &Geschrieben, NULL);
	res = ReadFile(hCom, IStatus, 2, &Gelesen, NULL);
	if (Gelesen == 0) {return(ftiFehler);}

	Interface.InputStatus = unsigned char(IStatus[0]) + (unsigned char(IStatus[1])<<8);
	return(Interface.InputStatus);
}

// GetInputsLPT -------------------------------------------------------------
//
// Setzen Motorstatus, Lesen Inputstatus
//
// --------------------------------------------------------------------------

DWORD GetInputsLPT(ftiDCB &Interface)
{
	DWORD i, IStatus = 0;

// ---- bertragen Motorstatus ---

	for (i=16; i>=1; i--)
	{
		if(Interface.OutputStatus & EMaske[i])
		{	// --- Motor luft
			SetLPT(Interface, 0x34);              // --- DATA-OUT & Trigger
			SetLPT(Interface, 0x3C);              // --- DATA-OUT, CLOCK, Trigger
		}
		else
		{                                         // --- Motor steht
			SetLPT(Interface, 0x30);              // --- 0 & Trigger
			SetLPT(Interface, 0x38);              // --- CLOCK & Trigger
		}
	}
	SetLPT(Interface, 0x39);                      // --- LOAD-OUT, CLOCK, Trigger

// ---- Lesen Inputstatus -------

	SetLPT(Interface, 0x32);                     // --- LOAD-IN
	SetLPT(Interface, 0x3A);                     // --- LOAD & CLOCK
	for (i=8; i>=1; i--)                         // --- Erstes Interface
	{                                            // --- Auslesen einzelner Eingang
		if(GetLPT(Interface) & 0x80) IStatus |= EMaske[i];
		SetLPT(Interface, 0x30);                 // --- null
		SetLPT(Interface, 0x38);                 // --- CLOCK
	}

	for (i=16; i>=9; i--)                        // --- Slave Interface
	{                                            // --- Auslesen einzelner Eingang
		if(GetLPT(Interface) & 0x80) IStatus |= EMaske[i];
		SetLPT(Interface, 0x30);                 // --- null
		SetLPT(Interface, 0x38);                 // --- CLOCK
	}
  
	Interface.InputStatus = IStatus;
	return(IStatus);
}


// GetAnalog ----------------------------------------------------------------
// 
// Abfrage eines der beiden Analogeingnge
//
// ----------------------------------------------------------------------------

DWORD GetAnalog(ftiDCB &Interface, DWORD AnalogNr)
{
	if(Interface.hCom == 0) return(ftiFehler);
	if(Interface.PortID == 0)
		if(Interface.ECount == 8) return(GetAnalogCOM1(Interface, AnalogNr)); 
		else return(GetAnalogCOM2(Interface, AnalogNr));
	else
		if(Interface.PortID == 1) return(GetAnalogRT(Interface, AnalogNr));
		else return(GetAnalogLPT(Interface, AnalogNr));
}


// GetAnalogCOM1 ----------------------------------------------------------------
//
// Abfrage eines der beiden COM-Analogeingnge
//
// -----------------------------------------------------------------------------

DWORD GetAnalogCOM1(ftiDCB &Interface, DWORD AnalogNr)
{
	HANDLE hCom = Interface.hCom;
	unsigned char Befehl[2];
	unsigned char IAStatus[3];
	DWORD res, AnalogWert;
	DWORD Geschrieben, Gelesen;


	if(AnalogNr == 0) Befehl[0] = '\xC5'; else Befehl[0] = '\xC9';
	Befehl[1] = unsigned char(Interface.OutputStatus & 0xFF);

	res = WriteFile(hCom, Befehl, 2, &Geschrieben, NULL);
	res = ReadFile(hCom, IAStatus, 3, &Gelesen, NULL);
	if(Gelesen == 0) 
	{
		Interface.InputStatus  = ftiFehler; 
		return(ftiFehler);	
	}

	AnalogWert = (unsigned char(IAStatus[1])<<8) + unsigned char(IAStatus[2]);
	if (AnalogWert < 0 || AnalogWert > 2048) AnalogWert = 2048;
	Interface.InputStatus = IAStatus[0];
	return (AnalogWert);
}

// GetAnalogCOM2 ----------------------------------------------------------------
//
// Abfrage eines der beiden COM-Analogeingnge
//
// -----------------------------------------------------------------------------

DWORD GetAnalogCOM2(ftiDCB &Interface, DWORD AnalogNr)
{
	HANDLE hCom = Interface.hCom;
	unsigned char Befehl[3];
	unsigned char IAStatus[4];
	DWORD res, AnalogWert;
	DWORD Geschrieben, Gelesen;


	if(AnalogNr == 0) Befehl[0] = '\xC6'; else Befehl[0] = '\xCA';
	Befehl[1] = unsigned char(Interface.OutputStatus & 0xFF);
	Befehl[2] = unsigned char((Interface.OutputStatus & 0xFF00)>>8);

	res = WriteFile(hCom, Befehl, 3, &Geschrieben, NULL);
	res = ReadFile(hCom, IAStatus, 4, &Gelesen, NULL);
	if(Gelesen == 0) 
	{
		Interface.InputStatus  = ftiFehler; 
		return(ftiFehler);	
	}

	AnalogWert = (unsigned char(IAStatus[2])<<8) + unsigned(IAStatus[3]);
	if (AnalogWert < 0 || AnalogWert > 2048) AnalogWert = 2048;
	Interface.InputStatus = unsigned char(IAStatus[0]) + (unsigned char(IAStatus[1])<<8);;
	return (AnalogWert);
}

// GetAnalogLPT ----------------------------------------------------------------
//
// Abfrage eines der beiden LPT-Analogeingnge
//
// -----------------------------------------------------------------------------

DWORD GetAnalogLPT(ftiDCB &Interface, DWORD AnalogNr)
{
	DWORD i, AWert = 0;

// --- Trigger X bzw. Y ----
	if(AnalogNr == 0) SetLPT(Interface, 0xA0); else SetLPT(Interface, 0x90);
	SetLPT(Interface, 0x38);              // --- CLOCK

// --- Zhlschleife ----

	i = 0;
	do 
	{	// --- Loop bei Input = 0xF8
		i++;
		if(i == Interface.LPTAnalog) 
		{
			AWert += 1;
			i = 0;
			if(AWert >= 2048) break;
		}
	}
	while (GetLPT(Interface) & 0x80);     // --- Ende bei Input = 0x80
  
	return(AWert);
}



// umGetAnalogDirect -------------------------------------------------------------
// 
// Lesen Analogwert AnalogNr direkt. Dazu wird das Pollen kurzzeitig ausgeschaltet
//
// -------------------------------------------------------------------------------

DWORD __stdcall umGetAnalogDirect(ftiDCB &Interface, DWORD AnalogNr)
{
	DWORD res;

	res = ftiFehler;
	if (Interface.hCom == 0) return(res);

	// --- Pollen Stoppen -----------------------------------------
	Sleep(2*Interface.PollInterval);
	if(Interface.FID != 0) timeKillEvent(Interface.FID);

	// --- Analogwert einlesen ------------------------------------
	res = GetAnalog(Interface, AnalogNr);

	// --- Pollen wieder starten ----------------------------------
	Interface.FID = timeSetEvent(
		Interface.PollInterval,
		0,                 // --- Auflsung in mSek (0 = max)
		PollInterface,     // --- CallBack-Routine
		DWORD(&Interface), // --- bergabe ftiDCB
		TIME_PERIODIC);

	return res;
}


// umGetInput -------------------------------------------------------------------
// 
// Lesen Status Eingang InputNr
//
// -----------------------------------------------------------------------------

BOOL __stdcall umGetInput(ftiDCB &Interface, DWORD InputNr)
{
	DWORD res;

	res = ftiFehler;
	if (Interface.hCom == 0) return(res);
	
	res = Interface.InputStatus;
	return((res & EMaske[InputNr]) ? -1 : 0);
}


// SetMotorAll ----------------------------------------------------------------
//
// Schalten des (Motor)Ausganges MotorNr unter Angabe 
// von Richtung, Geschwindigkeit und robMode
//
// -----------------------------------------------------------------------------

DWORD __stdcall SetMotorAll(ftiDCB &Interface, DWORD MotorNr, DWORD Direction, 
						    DWORD Speed, DWORD RobMode, int ICount)
{
	DWORD MotorStatus, SpeedStatus, ModeStatus, res;

	res = ftiFehler;
	if (Interface.hCom == 0) return(res);

	MotorStatus = Interface.MotorStatus & MAus[MotorNr];
	SpeedStatus = Interface.SpeedStatus & ~SMaske[MotorNr];
	ModeStatus  = Interface.ModeStatus  & ~SMaske[MotorNr];
	switch (Direction) 
	{
		case ftiAus: break;
		case ftiLinks:  MotorStatus |= MLinks[MotorNr]; break;
		case ftiRechts: MotorStatus |= MRechts[MotorNr]; break;
		default: MotorStatus = 0;
	}
	Interface.OutputStatus = MotorStatus;
	Interface.MotorStatus  = MotorStatus;
	Interface.SpeedStatus  = SpeedStatus | (Speed   << (MotorNr-1)*4);
	Interface.ModeStatus   = ModeStatus  | (RobMode << (MotorNr-1)*4);
	if(ICount >= 0) Interface.Counters[(MotorNr-1)*2+1] = ICount;
	res = Interface.InputStatus;
	return(res);
}

// umSetMotor --------------------------------------------------------------------
//
// Schalten des (Motor)Ausganges MotorNr
//
// -----------------------------------------------------------------------------

DWORD __stdcall umSetMotor(ftiDCB &Interface, DWORD MotorNr, DWORD Direction)
{
	return SetMotorAll(Interface, MotorNr, Direction, 0xF, 0, -1);
}

// umSetMotorEx ----------------------------------------------------------------
//
// Schalten des (Motor)Ausganges MotorNr unter Angabe von Richtung und Geschw.
//
// -----------------------------------------------------------------------------

DWORD __stdcall umSetMotorEx(ftiDCB &Interface, DWORD MotorNr, DWORD Direction, DWORD Speed)
{
	return SetMotorAll(Interface, MotorNr, Direction, Speed, 0, -1);
}

// umSetMotors ----------------------------------------------------------------
//
// Schalten aller M-Ausgnge, Speed = 15.
//
// -----------------------------------------------------------------------------

DWORD __stdcall umSetMotors(ftiDCB &Interface, DWORD MotorStatus)
{
	Interface.OutputStatus = MotorStatus;
	Interface.MotorStatus  = MotorStatus;
	Interface.SpeedStatus  = 0xFFFFFFFF;
	Interface.ModeStatus   = 0;
	return Interface.InputStatus;
}

// umSetMotorsEx --------------------------------------------------------------
//
// Schalten aller M-Ausgnge, einschl. Geschwindigkeit
//
// -----------------------------------------------------------------------------

DWORD __stdcall umSetMotorsEx(ftiDCB &Interface, DWORD MotorStatus, DWORD SpeedStatus)
{
	Interface.OutputStatus = MotorStatus;
	Interface.MotorStatus  = MotorStatus;
	Interface.SpeedStatus  = SpeedStatus;
	Interface.ModeStatus   = 0;
	return Interface.InputStatus;
}


// umSetLamp --------------------------------------------------------------------
//
// Schalten des LampenAusganges LampNr 1 - 8 (16)
//
// -----------------------------------------------------------------------------

DWORD __stdcall umSetLamp(ftiDCB &Interface, DWORD LampNr, DWORD OnOff)
{
	DWORD res;
	res = ftiFehler;
	if (Interface.hCom == 0) return(res);

	if(Interface.PortID)       // --- paralleles Interface, PortID != 0
	{
		Interface.OutputStatus &= ~EMaske[LampNr];   // --- Lschen LampNr
		if(OnOff) Interface.OutputStatus |= EMaske[LampNr];
	}
	else                              // --- serielles Interface (PortID = 0)
	{
		Interface.OutputStatus |= EMaske[LampNr];
		if(OnOff) Interface.OutputStatus &= ~EMaske[LampNr];
	}
	Interface.MotorStatus  = Interface.OutputStatus;
	Interface.SpeedStatus |= (0x3 << 2*(LampNr-1));
	res = Interface.InputStatus;
	return(res);
}


// umRobMotor ------------------------------------------------------------------
//
// Schalten des M-Ausganges MotorNr unter ModeRob-Bedingungen
// Motor wird bei zugehrenden ImpulsCounter = 0 oder EndTaster abgeschaltet
// zu Motor m gehren EndTaster (m-1)*2+1 und ImpulsTaster/Counter (m-1)*2+2
//
// -----------------------------------------------------------------------------

DWORD __stdcall umRobMotor(ftiDCB &Interface, DWORD MotorNr, DWORD Direction, 
						   DWORD Speed, DWORD ICount)
{
	return SetMotorAll(Interface, MotorNr, Direction, Speed, 1, ICount);
}

// umRobMotors -----------------------------------------------------------------
//
// Schalten aller M-Ausgnge unter ModeRob-Bedingungen
// Motoren werden bei zugehrenden ImpulsCounter = 0 oder EndTaster 
// selektiv abgeschaltet
// zu Motor m gehren EndTaster (m-1)*2+1 und ImpulsTaster/Counter (m-1)*2+2
//
// -----------------------------------------------------------------------------

DWORD __stdcall umRobMotors(ftiDCB &Interface, DWORD MotorStatus, DWORD SpeedStatus,
							DWORD ModeStatus)
{
	Interface.OutputStatus = MotorStatus;
	Interface.MotorStatus  = MotorStatus;
	Interface.SpeedStatus  = SpeedStatus;
	Interface.ModeStatus   = ModeStatus;
	return Interface.InputStatus;
}

// ----------------------------------------------------------------------------------
// --------------------- cs-Variante ------------------------------------------------
// ----------------------------------------------------------------------------------


int __stdcall csOpenInterfaceEx(int PortNr, int AnalogScan, int Slave, int PollInterval, 
								int LPTAnalog, int LPTDelay)
{
	int    iHandle = ftiFehler;
	LPCSTR PortName;
	switch(PortNr)
	{
	case 0:  PortName  = "LPT";  break;
	case 1:  PortName  = "COM1"; break;
	case 2:  PortName  = "COM2"; break;
	case 3:  PortName  = "COM3"; break;
	case 4:  PortName  = "COM4"; break;
	case 5:  PortName  = "COM5"; break;
	case 6:  PortName  = "COM6"; break;
	case 7:  PortName  = "COM7"; break;
	case 8:  PortName  = "COM8"; break;
	case 9:  PortName  = "LPT1"; break;
	case 10: PortName  = "LPT2"; break;
	case 11: PortName  = "LPT3"; break;
	default: PortName  = "COM1"; break;
	}
	for(int i=0; i<8; i++) {if(DCBPool[i] == 0){iHandle = i; break;}}
	if(iHandle == ftiFehler) return ftiFehler;
	DCBPool[iHandle] = 1;
	
	ftDCB[iHandle].LPTDelay     = LPTDelay;
	ftDCB[iHandle].LPTAnalog    = LPTAnalog;
	ftDCB[iHandle].ECount       = (Slave == 1) ? 16 : 8;
	ftDCB[iHandle].PollInterval = PollInterval;
	ftDCB[iHandle].AnalogScan   = AnalogScan;

	if(umOpenInterface(ftDCB[iHandle], PortName) == ftiFehler) return ftiFehler; 
	else return iHandle;
}

// csOpenInterface -------------------------------------------------------------
//
// Setzen Ausgangswerte, OpenSchnittstelle : Overload umOpenInterface
//
// -----------------------------------------------------------------------------

int __stdcall csOpenInterface(int PortNr, int AnalogScan, int Slave, int PollInterval)
{
	return csOpenInterfaceEx(PortNr, AnalogScan, Slave, PollInterval, 0 ,0);
}

// csOpenInterfaceEx -----------------------------------------------------------
//
// Setzen Ausgangswerte, OpenSchnittstelle : Overload umOpenInterface
//
// -----------------------------------------------------------------------------

// csCloseInterface ------------------------------------------------------------
//
// Schlieen der Schnittstelle : Overload umCloseInerface
//
// -----------------------------------------------------------------------------

int __stdcall csCloseInterface(int iHandle)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	DCBPool[iHandle] = 0;
	return (umCloseInterface(ftDCB[iHandle]));
}


// csVersion -------------------------------------------------------------------
//
// Ausgabe der eigenen Version
//
// -----------------------------------------------------------------------------

int __stdcall csVersion()
{
	return FVersion;
}


// csGetAnalogScan -------------------------------------------------------------
//
// Ausgabe mit/ohne AnalogScan : ftDCB
//
// -----------------------------------------------------------------------------

int __stdcall csGetAnalogScan(int iHandle)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return (ftDCB[iHandle].AnalogScan == 0) ? 0 : 1;
}


// csGetLPTAnalog ------------------------------------------------------------------
//
// Ausgabe Analogwert-Skalierung : ftDCB
//
// -----------------------------------------------------------------------------

int __stdcall csGetLPTAnalog(int iHandle)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return ftDCB[iHandle].LPTAnalog;
}


// csGetLPTDelay ------------------------------------------------------------------
//
// Ausgabe Ausgabeverzgerung LPT-Port : ftDCB
//
// -----------------------------------------------------------------------------

int __stdcall csGetLPTDelay(int iHandle)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return ftDCB[iHandle].LPTDelay;
}


// csGetSlave ------------------------------------------------------------------
//
// Ausgabe mit/ohne Slave : ftDCB
//
// -----------------------------------------------------------------------------

int __stdcall csGetSlave(int iHandle)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return (ftDCB[iHandle].ECount == 16) ? 1 : 0;
}


// csGetPollInterval -----------------------------------------------------------
//
// Ausgabe Wert PollInterval : ftDCB
//
// -----------------------------------------------------------------------------

int __stdcall csGetPollInterval(int iHandle)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return ftDCB[iHandle].PollInterval; 
}


// csGetAnalogDirect -----------------------------------------------------------
//
// Lesen AnalogWert direkt : Overload umGetAnalogDirect
//
// -----------------------------------------------------------------------------

int __stdcall csGetAnalogDirect(int iHandle, int AnalogNr)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return umGetAnalogDirect(ftDCB[iHandle], AnalogNr);
}


// csGetInput ------------------------------------------------------------------
//
// Ausgabe Status E-Eingang : Overload umGetInput
//
// -----------------------------------------------------------------------------

int __stdcall csGetInput(int iHandle, int InputNr)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return ((umGetInput(ftDCB[iHandle], InputNr) == 0) ? 0 : 1);
}


// csGetInputs -----------------------------------------------------------------
//
// Ausgabe Status der E-Eingnge : ftDCB
//
// -----------------------------------------------------------------------------

int __stdcall csGetInputs(int iHandle)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return ftDCB[iHandle].InputStatus;
}


// csGetAnalog -----------------------------------------------------------------
//
// Ausgabe Status A-Ausgang : ftDCB
//
// -----------------------------------------------------------------------------

int __stdcall csGetAnalog(int iHandle, int AnalogNr)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return ftDCB[iHandle].Analogs[AnalogNr];
}


// csSetMotor ------------------------------------------------------------------
//
// Setzen Status M-Ausgang : Overload umSetMotor
//
// -----------------------------------------------------------------------------

int __stdcall csSetMotor(int iHandle, int MotorNr, int Direction)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return SetMotorAll(ftDCB[iHandle], MotorNr, Direction, 0xF, 0, -1);
}


// csSetMotorEx ----------------------------------------------------------------
//
// Setzen Status M-Ausgang : Overload umSetMotor
//
// -----------------------------------------------------------------------------

int __stdcall csSetMotorEx(int iHandle, int MotorNr, int Direction, int Speed)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return SetMotorAll(ftDCB[iHandle], MotorNr, Direction, Speed, 0, -1);
}


// csGetMotors -----------------------------------------------------------------
//
// Ausgabe Status M-Ausgnge : ftDCB
//
// -----------------------------------------------------------------------------

int __stdcall csGetMotors(int iHandle)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return ftDCB[iHandle].MotorStatus;
}


// csSetMotors -----------------------------------------------------------------
//
// Setzen Status M-Ausgnge : ftDCB
//
// -----------------------------------------------------------------------------

int __stdcall csSetMotors(int iHandle, int MotorStatus)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return umSetMotors(ftDCB[iHandle], MotorStatus);
}


// csSetMotorsEx ---------------------------------------------------------------
//
// Setzen Status M-Ausgnge einschl. Speed : ftDCB
//
// -----------------------------------------------------------------------------

int __stdcall csSetMotorsEx(int iHandle, int MotorStatus, int SpeedStatus)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return umSetMotorsEx(ftDCB[iHandle], MotorStatus, SpeedStatus);
}


// csSetLamp -------------------------------------------------------------------
//
// Setzen Status halber M-Ausgang : OverLoad umSetLamp
//
// -----------------------------------------------------------------------------

int __stdcall csSetLamp(int iHandle, int LampNr, int OnOff)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return umSetLamp(ftDCB[iHandle], LampNr, OnOff);
}


// csRobMotor -------------------------------------------------------------------
//
// Starten Motor und zug. Counter : OverLoad umRobMotor
//
// -----------------------------------------------------------------------------

int __stdcall csRobMotor(int iHandle, int MotorNr, int Direction, 
						 int Speed, int ICount)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return SetMotorAll(ftDCB[iHandle], MotorNr, Direction, Speed, 1, ICount);
}


// csRobMotors -------------------------------------------------------------------
//
// Starten Motore und zugehrende Counters : OverLoad umRobMotors
//
// -----------------------------------------------------------------------------

int __stdcall csRobMotors(int iHandle, int MotorStatus, 
						  int SpeedStatus, int ModeStatus)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return umRobMotors(ftDCB[iHandle], MotorStatus, SpeedStatus, ModeStatus);
}


// csGetModeStatus ------------------------------------------------------------
//
// Lesen eines Betriebsmodes : ftDCB
//
// -----------------------------------------------------------------------------

int __stdcall csGetModeStatus(int iHandle, int MotorNr)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return (ftDCB[iHandle].ModeStatus & SMaske[MotorNr]) >> (MotorNr-1)*4;
}


// csSetModeStatus ------------------------------------------------------------
//
// Setzen eines Betriebsmodes : ftDCB
//
// -----------------------------------------------------------------------------

void __stdcall csSetModeStatus(int iHandle, int MotorNr, int Value)
{
	if(iHandle < 0 || iHandle > 7) return;
	ftDCB[iHandle].ModeStatus &= ~SMaske[MotorNr];
	ftDCB[iHandle].ModeStatus |= Value<<(MotorNr-1)*4;
}


// csGetCounter ----------------------------------------------------------------
//
// Ausgabe Wert eines Counters : ftDCB
//
// -----------------------------------------------------------------------------

int __stdcall csGetCounter(int iHandle, int CounterNr)
{
	if(iHandle < 0 || iHandle > 7) return ftiFehler;
	return ftDCB[iHandle].Counters[CounterNr-1];
}


// csSetCounter ----------------------------------------------------------------
//
// Setzen Wert eines Counters : ftDCB
//
// -----------------------------------------------------------------------------

void __stdcall csSetCounter(int iHandle, int CounterNr, int Value)
{
	if(iHandle < 0 || iHandle > 7) return;
	ftDCB[iHandle].Counters[CounterNr-1] = Value;
}


// csClearCounters -------------------------------------------------------------
//
// Lschen aller Counter : ftDCB
//
// -----------------------------------------------------------------------------

void __stdcall csClearCounters(int iHandle)
{
	if(iHandle < 0 || iHandle > 7) return;
	for(int i=0; i<16; i++) ftDCB[iHandle].Counters[i] = 0;
}


// GetLPT ----------------------------------------------------------------------
//
// Lesen eines WORD (2 Byte : Integer) vom aktuellen LPT-Port+1
//
// -----------------------------------------------------------------------------

DWORD GetLPT(ftiDCB &Interface)
{
  WORD res;
  WORD portid = WORD(Interface.PortID) + 1;

    __asm {
      push dx
      mov dx, portid
      in ax, dx
      mov res, ax
      pop dx
    }

  return(res);
}

// SetLPT       ----------------------------------------------------------------
//
// Schreiben eines WORD (2 Byte : Integer) auf den aktuellen LPT-Port
//
// -----------------------------------------------------------------------------

VOID  SetLPT(ftiDCB &Interface, WORD Wert)
{
  WORD portid = WORD(Interface.PortID);

  for (DWORD i=1; i<=Interface.LPTDelay; i++) 
  {
    __asm {
      push dx
      mov dx, portid
      mov ax, Wert
      out dx, ax
      pop dx
    } 
  }

}



// DllMain ---------------------------------------------------------------------------------
//
// Hauptschleife bei Start und Ende
//
// -----------------------------------------------------------------------------------------

BOOL WINAPI DllMain (HINSTANCE hDLL, DWORD dwReason, LPVOID)
{
	switch(dwReason) {
		case DLL_PROCESS_DETACH:
			umCloseInterface(ftiSave);
			break;
	}
	return TRUE;
}